@@ -1,17 +1,17 @@
 ;;; inform-mode.el --- Inform mode for Emacs
 
 ;; Original-Author: Gareth Rees <Gareth.Rees@cl.cam.ac.uk>
-;; Maintainer: Rupert Lane <rupert@merguez.demon.co.uk>
+;; Maintainer: Rupert Lane <rupert@rupert-lane.org>
 ;; Created: 1 Dec 1994
-;; Version: 1.5.6
-;; Released: 17 Dec 2000
+;; Version: 1.5.8
+;; Released: 3 Sep 2002
 ;; Keywords: languages
 
 ;;; Copyright:
 
 ;; Copyright (c) by Gareth Rees 1996
 ;; Portions copyright (c) by Michael Fessler 1997-1998
-;; Portions copyright (c) by Rupert Lane 1999-2000
+;; Portions copyright (c) by Rupert Lane 1999-2000, 2002
 
 ;; inform-mode is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -27,7 +27,7 @@
 
 ;; Inform is a compiler for adventure games by Graham Nelson,
 ;; available by anonymous FTP from
-;; /ftp.gmd.de:/if-archive/programming/inform/
+;; /ftp.ifarchive.org:/if-archive/programming/inform/
 ;;
 ;; This file implements a major mode for editing Inform programs.  It
 ;; understands most Inform syntax and is capable of indenting lines
@@ -53,7 +53,13 @@
 ;; To turn on font locking add:
 ;; (add-hook 'inform-mode-hook 'turn-on-font-lock)
 
-;; Please send any bugs or comments to rupert@merguez.demon.co.uk
+;; If you use XEmacs and intend to use `inform-run-project' with a
+;; console-mode interpreter, you need to have the eterm package
+;; installed.  It should already be installed if you use XEmacs < 21,
+;; but starting with XEmacs 21.1 you may need to download and install
+;; it separately.
+
+;; Please send any bugs or comments to rupert@rupert-lane.org
 
 ;;; Code:
 
@@ -65,7 +71,7 @@
 ;;; General variables
 ;;;
 
-(defconst inform-mode-version "1.5.6")
+(defconst inform-mode-version "1.5.8")
 
 (defvar inform-maybe-other 'c-mode
   "*`inform-maybe-mode' runs this if current file is not in Inform mode.")
@@ -89,8 +95,8 @@ If you do not want a leading newline before opening braces then use:
     (define-key inform-mode-map "\177" 'backward-delete-char-untabify)
     (define-key inform-mode-map "\C-c\C-r" 'inform-retagify)
     (define-key inform-mode-map "\C-c\C-t" 'visit-tags-table)
-    (define-key inform-mode-map "\C-c\C-o" 'inform-convert-old-format)
     (define-key inform-mode-map "\C-c\C-b" 'inform-build-project)
+    (define-key inform-mode-map "\C-c\C-c" 'inform-run-project)
     (define-key inform-mode-map "\C-c\C-a" 'inform-toggle-auto-newline)
     (define-key inform-mode-map "\C-c\C-s" 'inform-spell-check-buffer)
     (define-key inform-mode-map "\M-n" 'inform-next-object)
@@ -104,8 +110,6 @@ If you do not want a leading newline before opening braces then use:
     (define-key inform-mode-map "," 'inform-electric-comma)
     (define-key inform-mode-map [menu-bar] (make-sparse-keymap))
     (define-key inform-mode-map [menu-bar inform] (cons "Inform" map))
-    (define-key map [convert] 
-      '("Convert old format" . inform-convert-old-format))
     (define-key map [separator4] '("--" . nil))
     (define-key map [inform-spell-check-buffer] 
       '("Spellcheck buffer" . inform-spell-check-buffer))
@@ -115,6 +119,7 @@ If you do not want a leading newline before opening braces then use:
     (define-key map [load-tags] '("Load tags table" . visit-tags-table))
     (define-key map [retagify] '("Rebuild tags table" . inform-retagify))
     (define-key map [build] '("Build project" . inform-build-project))
+    (define-key map [run] '("Run project" . inform-run-project))
     (define-key map [separator2] '("--" . nil))
     (define-key map [next-object] '("Next object" . inform-next-object))
     (define-key map [prev-object] '("Previous object" . inform-prev-object))
@@ -149,6 +154,24 @@ If you do not want a leading newline before opening braces then use:
 (defvar inform-command-options ""
   "*Options with which to call the Inform compiler.")
 
+(defvar inform-interpreter-command "frotz"
+  "*The command with which to run the ZCode interpreter.
+If a string, the name of a command.  If a symbol or a function value, an
+Emacs-lisp function to be called with the name of the story file.")
+
+(defvar inform-interpreter-options ""
+  "*Additional options with which to call the ZCode interpreter.
+Only used if `inform-interpreter-command' is a string.")
+
+(defvar inform-interpreter-kill-old-process t
+  "*Whether to kill the old interpreter process when starting a new one.")
+
+(defvar inform-interpreter-is-graphical nil
+  "*Controls whether `inform-interpreter-command' will be run in a buffer.
+If NIL, `inform-run-project' will switch to the interpreter buffer after
+running the interpreter.")
+
+
 
 ;;;
 ;;; Indentation parameters
@@ -240,6 +263,19 @@ first line.")
 Used to build a font-lock regexp; the name defined must follow the
 keyword.")
 
+  ;; We have to hardcode the regexp for inform-defining-list due to the way
+  ;; regexp-opt works on different emacsen.
+  ;; On Emacs 20 it always uses regular \( \) grouping
+  ;; On Emacs 21 it always uses shy \(?: \) grouping
+  ;; On XEmacs it can use either based on the shy parameter.
+  ;; This means it is impossible to write a match-string expression in
+  ;; inform-font-lock-keywords using regexp-opt that will work on all emacsen.
+  ;; If Emacs 20 support is dropped this should be removed and shy grouping
+  ;; used.
+  (defvar inform-defining-list-regexp
+    "\\[\\|a\\(rray\\|ttribute\\)\\|c\\(lass\\|onstant\\)\\|fake_action\\|global\\|lowstring\\|nearby\\|object\\|property"
+    "Regexp based on inform-defining-list, hardcoded for portability.")
+
   (defvar inform-attribute-list
     '("absent" "animate" "clothing" "concealed" "container" "door"
       "edible" "enterable" "female" "general" "light" "lockable" "locked"
@@ -340,17 +376,18 @@ That is, one found at the start of a line.")
     (list
 
      ;; Inform code keywords
-     (cons (concat "\\s-+\\("
+     (cons (concat "\\s-\\("
                    (inform-make-regexp inform-code-keyword-list)
                    "\\)\\(\\s-\\|$\\|;\\)")
-           'font-lock-keyword-face)
+           '(1 font-lock-keyword-face))
      
      ;; Keywords that declare variable or constant names.
-     (list (concat "^#?\\("
-                   (inform-make-regexp inform-defining-list nil t)
-		   "\\)\\s-+\\(->\\s-+\\)*\\(\\(\\w\\|\\s_\\)+\\)")
-           '(1 font-lock-keyword-face)
-           '(5 font-lock-function-name-face))
+     (list 
+      (concat "^#?\\("
+              inform-defining-list-regexp
+              "\\)\\s-+\\(->\\s-+\\)*\\(\\(\\w\\|\\s_\\)+\\)")
+      '(1 font-lock-keyword-face)
+      '(5 font-lock-function-name-face))
 
      ;; Other directives.
      (cons inform-directive-regexp 'font-lock-keyword-face)
@@ -359,8 +396,15 @@ That is, one found at the start of a line.")
      '("'\\(\\(-\\|\\w\\)\\(\\(-\\|\\w\\)+\\(//\\w*\\)?\\|//\\w*\\)\\)'"
        (1 inform-dictionary-word-face append))
 
-     ;; Also quoted words after "name" property
-     '("\\s-name\\s-+"
+     ;; Double-quoted dictionary words
+     '("\\(\\s-name\\s-\\|^Verb\\|^Extend\\|^\\s-+\\*\\)"
+       ("\"\\(\\(-\\|\\w\\)+\\)\"" nil nil
+        (1 inform-dictionary-word-face t)))
+
+     ;; More double-quoted dictionary words
+     '("^\\s-+\"\\(\\(-\\|\\w\\)+\\)\"\\s-+\"\\(\\(-\\|\\w\\)+\\)\""
+       (1 inform-dictionary-word-face t)
+       (3 inform-dictionary-word-face t)
        ("\"\\(\\(-\\|\\w\\)+\\)\"" nil nil
         (1 inform-dictionary-word-face t)))
 
@@ -419,6 +463,11 @@ That is, one found at the start of a line.")
   Type \\[inform-build-project] to build the current project.
   Type \\[next-error] to go to the next error.
 
+* Running:
+
+  Type \\[inform-run-project] to run the current project in an
+  interpreter, either as a sepaarte process or in an Emacs terminal buffer.
+
 * Font-lock support:
 
   Put \(add-hook 'inform-mode-hook 'turn-on-font-lock) in your .emacs.
@@ -428,12 +477,6 @@ That is, one found at the start of a line.")
   Type \\[inform-spell-check-buffer] to spell check all strings in the buffer.
   Type \\[ispell-word] to check the single word at point.
 
-* Old versions of Inform Mode:
-
-  Versions of Inform Mode prior to 0.5 used tab stops every 4 characters
-  to control the formatting.  This was the Wrong Thing To Do.
-  Type \\[inform-convert-old-format] to undo the broken formatting.
-
 * Key definitions:
 
 \\{inform-mode-map}
@@ -520,7 +563,29 @@ That is, one found at the start of a line.")
   inform-command-options
     Additional options with which to call the Inform compiler.
 
-* Please send any bugs or comments to rupert@merguez.demon.co.uk
+* User options to do with an interpreter:
+
+  inform-interpreter-command
+    The command with which to run the ZCode interpreter.  Can be a
+    string (a command to be run), a symbol (name of function to call)
+    or a function.
+
+  inform-interpreter-options
+    Additional options with which to call the ZCode interpreter.  Only
+    used if `inform-interpreter-command' is a string.
+
+  inform-interpreter-kill-old-process
+    If non-NIL, `inform-run-project' will kill any running interpreter
+    process and start a new one.  If not, will switch to the interpreter's
+    buffer (if necessary - see documentation for `inform-run-project' for
+    details).
+
+  inform-interpreter-is-graphical
+    If NIL, `inform-run-project' will switch to the interpreter buffer
+    after running the interpreter.
+
+
+* Please send any bugs or comments to rupert@rupert-lane.org
 "
 
   (interactive)
@@ -1040,16 +1105,16 @@ comments, return CURRENT-INDENT."
       (inform-calculate-indentation (inform-syntax-class))
     (max (1+ (current-column)) comment-column)))
 
-;; Indent line containing point.  If the indentation changes or if point
-;; is before the first non-whitespace character on the line,
-;; move point to indentation.
+;; Indent line containing point. 
+;; Keep point at the "logically" same place, unless point was before
+;; new indentation, in which case place point at indentation.
 
 (defun inform-indent-line ()
-  (let ((old-point (point)))
+  (let ((oldpos (- (point-max) (point))))
     (forward-line 0)
-    (or (inform-do-indent-line (inform-syntax-class))
-        (< old-point (point))
-        (goto-char old-point))))
+    (inform-do-indent-line (inform-syntax-class))
+    (and (< oldpos (- (point-max) (point)))
+         (goto-char (- (point-max) oldpos)))))
 
 ;; Indent all the lines in region.
 
@@ -1151,7 +1216,7 @@ comments, return CURRENT-INDENT."
 ;; loops.
 
 (defun inform-project-file-list ()
-  (let* ((project-file (inform-project-file))
+  (let* ((project-file (expand-file-name (inform-project-file)))
          (project-dir (file-name-directory project-file))
          (in-file-list (list project-file))
          out-file-list
@@ -1207,12 +1272,13 @@ table."
 
     ;; Uses call-process to work on windows/nt systems (not tested)
     ;; Regexp matches routines or object/class definitions
-    (call-process inform-etags-program
-                  nil nil nil
-                  "--regex=/\\([oO]bject\\|[nN]earby\\|[cC]lass\\|\\[\\)\\([ \\t]*->\\)*[ \\t]*\\([A-Za-z0-9_]+\\)/"
-                  (concat "--output=" tags-file)
-                  "--language=none"
-                  (mapconcat (function (lambda (x) x)) files ""))
+    (apply (function call-process)
+           inform-etags-program
+           nil nil nil
+           "--regex=/\\([oO]bject\\|[nN]earby\\|[cC]lass\\|\\[\\)\\([ \\t]*->\\)*[ \\t]*\\([A-Za-z0-9_]+\\)/"
+           (concat "--output=" tags-file)
+           "--language=none"
+           files)
         
     (message "Running external tags program...done")
     (inform-auto-load-tags-table)))
@@ -1350,6 +1416,11 @@ With a negative prefix arg, go forwards."
               (buffer-substring-no-properties (match-beginning 3)
                                               (match-end 3)))))
 
+
+;;;
+;;; Build and run project
+;;;
+
 (defun inform-build-project ()
   "Compile the current Inform project.
 The current project is given by `inform-project-file', or the current
@@ -1370,14 +1441,71 @@ file if this is NIL."
                  (substring project-file 0 (match-beginning 1))
                project-file)))))
 
-(defun inform-convert-old-format ()
-  "Undoes Inform Mode 0.4 formatting for current buffer.
-Early versions of Inform Mode used tab stops every 4 characters to
-control formatting, which was the Wrong Thing To Do.  This function
-undoes that stupidity."
+(defun inform-run-project ()
+  "Run the current Inform project using `inform-interpreter-command'.
+The current project is given by `inform-project-file', or the current
+file if this is NIL.  Will kill any running interpreter if
+`inform-interpreter-kill-old-process' is non-NIL.  Switches to the
+interpreter's output buffer if `inform-interpreter-is-graphical' is
+NIL."
   (interactive)
-  (let ((tab-width 4))
-    (untabify (point-min) (point-max))))
+  (let* ((project-file (inform-project-file))
+         (story-file-base (if (string-match "\\`[^.]+\\(\\.inf\\'\\)"
+                                            project-file)
+                              (substring project-file 0 (match-beginning 1))
+                            project-file))
+         (story-file (concat story-file-base
+                             (if (string-match "-v8" inform-command-options)
+                                 ".z8"
+                               ".z5")))
+         (name "Inform interpreter"))
+    (if  (or (symbolp inform-interpreter-command)
+             (functionp inform-interpreter-command))
+        ;; Emacs interpreter (or custom function)
+        (funcall inform-interpreter-command story-file)
+      ;; inform-interpreter-command is truly a command
+      (let* ((buffer (get-buffer-create (concat "*" name "*")))
+             (proc (get-buffer-process buffer)))
+        (and inform-interpreter-kill-old-process
+             proc
+             (kill-process proc))
+        (if (or inform-interpreter-is-graphical
+                (eq window-system 'w32)) ; Windows can't handle
+                                        ; term-exec anyway
+            (progn
+              ;; X gets confused if an application is restarted too quickly
+              ;; Assume X if not Win32
+              (unless (eq window-system 'w32)
+                (message "Waiting for X...")
+                ;; On my system 0.1 seconds was enough - double it for safety
+                (sleep-for 0.2)
+                (message ""))
+              (when (or inform-interpreter-kill-old-process
+                        (not proc))
+                (apply (function start-process)
+                       name buffer inform-interpreter-command
+                       ;; Some shells barf on "empty" arguments
+                       (if (string-equal "" inform-interpreter-options)
+                           (list story-file)
+                         (list inform-interpreter-options
+                               story-file)))))
+          ;; Console-mode 'terp
+          (require 'term)
+          (when (or inform-interpreter-kill-old-process
+                    (not proc))
+            (set-buffer buffer)
+            (term-mode)
+            (erase-buffer)
+            (term-exec buffer name inform-interpreter-command nil
+                       (if (string-equal "" inform-interpreter-options)
+                           (list story-file)
+                         (list inform-interpreter-options
+                               story-file)))
+            (term-char-mode)
+            (term-pager-disable))
+          (switch-to-buffer buffer)
+          (goto-char (point-max)))))))
+
 
 
 ;;;
